# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Revisions response models for the Store."""

from ._base_model import MarshableModel


class RevisionsRequestModel(MarshableModel):
    """Resource model for a ReleaseRequestModel.

    :param upload_id: the upload-id returned from the storage endpoint.
    """

    upload_id: str


class RevisionsResponseModel(MarshableModel):
    """Model for a revisions response.

    :param status-url: a URL to monitor the state of the posted revision.
    """

    status_url: str



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Attenuations for Snap Store and Charmhub discharged Macaroons."""

# read/write access.
ACCOUNT_REGISTER_PACKAGE = "account-register-package"
"""Register or request a new package name under a given account."""

PACKAGE_MANAGE = "package-manage"
"""
Meta permission for easing creation of a complete RW token,
it grants all the package-manage-* permissions.
"""

PACKAGE_MANAGE_ACL = "package-manage-acl"
"""Add, invite or remove collaborators."""

PACKAGE_MANAGE_LIBRARY = "package-manage-library"
"""Register or request a new library name under a given package."""

PACKAGE_MANAGE_METADATA = "package-manage-metadata"
"""Edit metadata, add or remove media, etc."""

PACKAGE_MANAGE_RELEASES = "package-manage-releases"
"""
Release revisions, close channels and update version pattern for a track.
"""

PACKAGE_MANAGE_REVISIONS = "package-manage-revisions"
"""
Upload new blobs, check for upload status, reject a revision blocked on
manual review or request manual review.
"""

# read only access.
ACCOUNT_VIEW_PACKAGES = "account-view-packages"
"""
List packages owned by the account and packages for which this account
has collaborator rights.
"""

PACKAGE_VIEW = "package-view"
"""
Meta permission for easing creation of a complete read only token grants all the
package-view-* permissions.
"""

PACKAGE_VIEW_ACL = "package-view-acl"
"""List the collaborators for a package and privacy settings."""

PACKAGE_VIEW_METADATA = "package-view-metadata"
"""View the metadata for a package, including media."""

PACKAGE_VIEW_METRICS = "package-view-metrics"
"""View the metrics of a package."""

PACKAGE_VIEW_RELEASES = "package-view-releases"
"""
List the current releases (channel map) for a package and the release
history of a package.
"""

PACKAGE_VIEW_REVISIONS = "package-view-revisions"
"""List the existing revisions for a package, along with status information."""



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""BaseModel with marshaling capabilities."""

from typing import Any, Dict

from pydantic import BaseModel


class MarshableModel(BaseModel):
    """A BaseModel that can be marshaled and unmarshaled."""

    class Config:  # type: ignore # pylint: disable=too-few-public-methods
        """Pydantic model configuration."""

        validate_assignment = True
        allow_mutation = False
        alias_generator = lambda s: s.replace("_", "-")  # noqa: E731

    @classmethod
    def unmarshal(cls, data: Dict[str, Any]) -> "MarshableModel":
        """Create and populate a new ``MarshableModel`` from a dict.

        The unmarshal method validates entries in the input dictionary, populating
        the corresponding fields in the data object.

        :param data: The dictionary data to unmarshal.

        :return: The newly created object.

        :raise TypeError: If data is not a dictionary.
        """
        if not isinstance(data, dict):
            raise TypeError("part data is not a dictionary")

        return cls(**data)

    def marshal(self) -> Dict[str, Any]:
        """Create a dictionary containing the part specification data.

        :return: The newly created dictionary.

        """
        return self.dict(by_alias=True)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Endpoint definitions for different services."""

import dataclasses
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Final, Optional, Sequence, Type

from overrides import overrides

from craft_store.models import (
    MarshableModel,
    charm_list_releases_model,
    snap_list_releases_model,
)


@dataclasses.dataclass(frozen=True)
class Package:
    """Representation of a package name and type."""

    package_name: str
    package_type: str


@dataclasses.dataclass(repr=True)
class Endpoints:  # pylint: disable=too-many-instance-attributes
    """Endpoints used to make requests to a store.

    :param namespace: the namespace to use for endpoints.
    :param whoami: path to the whoami API.
    :param upload: path used for uploading.
    :param tokens: path to the tokens API.
    :param tokens_exchange: path to the tokens_exchange API.
    :param tokens_refresh: path to the tokens_refresh API.
    :param list_releases_model: list_releases response model.
    """

    namespace: str
    whoami: str
    tokens: str
    tokens_exchange: str
    list_releases_model: Type[MarshableModel]
    valid_package_types: Sequence[str]
    upload: str = "/unscanned-upload/"
    tokens_refresh: Optional[str] = None

    def _validate_packages(self, packages: Sequence[Package]) -> None:
        unknown_packages = [
            p for p in packages if p.package_type not in self.valid_package_types
        ]
        if unknown_packages:
            unknown_package_types = [p.package_type for p in unknown_packages]
            raise ValueError(
                f"Package types {unknown_package_types} not in {self.valid_package_types}"
            )

    def get_token_request(
        self,
        *,
        permissions: Sequence[str],
        description: str,
        ttl: int,
        channels: Optional[Sequence[str]] = None,
        packages: Optional[Sequence[Package]] = None,
    ) -> Dict[str, Any]:
        """Return a properly formatted request for a token request.

        Permissions can be selected from :data:`craft_store.attenuations`

        :param permissions: a list of permissions to use.
        :param description: description that identifies the client.
        :param ttl: time to live for the requested token.
        :param packages: a sequence of :attr:`Package` to limit the requested token to.
        :param channels: a sequence of channels to limit the requested token to.
        """
        token_request = {
            "permissions": permissions,
            "description": description,
            "ttl": ttl,
        }

        if packages:
            self._validate_packages(packages)
            token_request["packages"] = [
                {"type": p.package_type, "name": p.package_name} for p in packages
            ]

        if channels:
            token_request["channels"] = channels

        return token_request

    @staticmethod
    def get_upload_id(result: Dict[str, Any]) -> str:
        """Return the upload ID for a given result.

        :param result: the result from an upload request.
        """
        return result["upload_id"]

    def get_releases_endpoint(self, name: str) -> str:
        """Return the slug to the releases endpoint."""
        return f"/v1/{self.namespace}/{name}/releases"

    def get_revisions_endpoint(self, name: str) -> str:
        """Return the slug to the revisions endpoint."""
        return f"/v1/{self.namespace}/{name}/revisions"


@dataclasses.dataclass(repr=True)
class _SnapStoreEndpoints(Endpoints):
    """Snap Store endpoints used to make requests to a store."""

    def get_token_request(
        self,
        *,
        permissions: Sequence[str],
        description: str,
        ttl: int,
        channels: Optional[Sequence[str]] = None,
        packages: Optional[Sequence[Package]] = None,
    ) -> Dict[str, Any]:
        expires = (
            datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc)
            + timedelta(seconds=ttl)
        ).isoformat()

        token_request: Dict[str, Any] = {
            "permissions": permissions,
            "description": description,
            "expires": expires,
        }

        if packages:
            self._validate_packages(packages)
            token_request["packages"] = [
                # Originally, snaps were supposed to be versioned by series,
                # this all changed with the introduction of bases.
                {"series": "16", "name": p.package_name}
                for p in packages
            ]

        if channels:
            token_request["channels"] = channels

        return token_request

    @staticmethod
    def get_upload_id(result: Dict[str, Any]) -> str:
        return result["upload_id"]

    @overrides
    def get_releases_endpoint(self, name: str) -> str:
        raise NotImplementedError()

    @overrides
    def get_revisions_endpoint(self, name: str) -> str:
        raise NotImplementedError()


CHARMHUB: Final = Endpoints(
    namespace="charm",
    whoami="/v1/tokens/whoami",
    tokens="/v1/tokens",
    tokens_exchange="/v1/tokens/exchange",
    valid_package_types=["charm", "bundle"],
    list_releases_model=charm_list_releases_model.ListReleasesModel,
)
"""Charmhub set of supported endpoints."""


SNAP_STORE: Final = _SnapStoreEndpoints(
    namespace="snap",
    whoami="/api/v2/tokens/whoami",
    tokens="/api/v2/tokens",
    tokens_exchange="/api/v2/tokens/exchange",
    valid_package_types=["snap"],
    list_releases_model=snap_list_releases_model.ListReleasesModel,
)
"""Snap Store set of supported endpoints."""


U1_SNAP_STORE: Final = _SnapStoreEndpoints(
    namespace="snap",
    whoami="/api/v2/tokens/whoami",
    tokens="/dev/api/acl/",
    tokens_exchange="/api/v2/tokens/discharge",
    tokens_refresh="/api/v2/tokens/refresh",
    valid_package_types=["snap"],
    list_releases_model=snap_list_releases_model.ListReleasesModel,
)
"""Ubuntu One compatible Snap Store set of supported endpoints."""



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Models package for store responses."""

from . import (
    charm_list_releases_model,
    release_request_model,
    revisions_model,
    snap_list_releases_model,
)
from ._base_model import MarshableModel
from .charm_list_releases_model import ListReleasesModel as CharmListReleasesModel
from .release_request_model import ReleaseRequestModel
from .revisions_model import RevisionsRequestModel, RevisionsResponseModel
from .snap_list_releases_model import ListReleasesModel as SnapListReleasesModel

__all__ = [
    "CharmListReleasesModel",
    "MarshableModel",
    "ReleaseRequestModel",
    "RevisionsRequestModel",
    "RevisionsResponseModel",
    "SnapListReleasesModel",
    "charm_list_releases_model",
    "release_request_model",
    "revisions_model",
    "snap_list_releases_model",
]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""List Releases Models for Snap Store responses."""


from datetime import datetime
from enum import Enum
from typing import List, Optional

from ._base_model import MarshableModel
from ._common_list_releases_model import PackageModel, ProgressiveModel


class ChannelMapModel(MarshableModel):
    """Model for the channel-map results from the list_releases endpoint."""

    architecture: str
    channel: str
    expiration_date: Optional[datetime]
    progressive: ProgressiveModel
    revision: int
    when: datetime


class Confinement(str, Enum):
    """Snap confinement."""

    CLASSIC = "classic"
    STRICT = "strict"
    DEVMODE = "devmode"


class Grade(str, Enum):
    """Snap grade."""

    DEVEL = "devel"
    STABLE = "stable"


class Type(str, Enum):
    """Type of snap."""

    APP = "app"
    GADGET = "gadget"
    KERNEL = "kernel"
    BASE = "base"
    SNAPD = "snapd"


class RevisionModel(MarshableModel):
    """Model for a revision entry from list_releases."""

    architectures: List[str]
    base: Optional[str]
    build_url: Optional[str]
    confinement: Confinement
    created_at: datetime
    created_by: str
    grade: Grade
    revision: int
    sha3_384: str
    size: int
    status: str
    type: Type
    version: str


class ListReleasesModel(MarshableModel):
    """Model for the list_releases endpoint."""

    channel_map: List[ChannelMapModel]
    package: PackageModel
    revisions: List[RevisionModel]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store HTTPClient."""

import logging
import os
from typing import Dict, Optional

import requests
from requests.adapters import HTTPAdapter, Retry

from . import errors

logger = logging.getLogger(__name__)


REQUEST_TOTAL_RETRIES = 8
"""Amount of retries for a request."""
REQUEST_BACKOFF = 1
"""Backoff before retrying a request."""


def _get_retry_value(environment_var: str, default_value: int) -> int:
    """Return the backoff to use in HTTPClient."""
    environment_value = os.getenv(environment_var)
    if environment_value is None:
        return default_value

    try:
        value = int(environment_value)
    except ValueError:
        logger.debug(
            "%r set to invalid value %r, setting to %r.",
            environment_var,
            environment_value,
            default_value,
        )
        return default_value

    if value < 0:
        logger.debug(
            "%r set to non positive value %r, setting to %r.",
            environment_var,
            value,
            default_value,
        )
        return default_value

    return value


class HTTPClient:
    """Generic HTTP Client to communicate with Canonical's Developer Gateway.

    This client has a requests like interface, it creates a requests.Session
    on initialization to handle retries over HTTP and HTTPS requests.

    The default number of retries is set in :data:`.REQUEST_TOTAL_RETRIES` and can
    be overridden with the ``CRAFT_STORE_RETRIES`` environment variable.

    The backoff factor has a default set in :data:`.REQUEST_BACKOFF` and can be
    overridden with the ``CRAFT_STORE_BACKOFF`` environment variable.

    Retries are done for the following return codes: ``500``, ``502``, ``503``
    and ``504``.

    :ivar user_agent: User-Agent header to identify the client.
    """

    def __init__(self, *, user_agent: str) -> None:
        """Initialize an HTTPClient with a given user_agent.

        :param user_agent: User-Agent header to identify the client.
        """
        self._session = requests.Session()
        self.user_agent = user_agent

        # Setup max retries for all store URLs and the CDN
        retries = Retry(
            total=_get_retry_value("CRAFT_STORE_RETRIES", REQUEST_TOTAL_RETRIES),
            backoff_factor=_get_retry_value("CRAFT_STORE_BACKOFF", REQUEST_BACKOFF),
            status_forcelist=[500, 502, 503, 504],
        )
        http_adapter = HTTPAdapter(max_retries=retries)
        self._session.mount("http://", http_adapter)
        self._session.mount("https://", http_adapter)

    def get(self, *args, **kwargs) -> requests.Response:
        """Perform an HTTP GET request."""
        return self.request("GET", *args, **kwargs)

    def post(self, *args, **kwargs) -> requests.Response:
        """Perform an HTTP POST request."""
        return self.request("POST", *args, **kwargs)

    def put(self, *args, **kwargs) -> requests.Response:
        """Perform an HTTP PUT request."""
        return self.request("PUT", *args, **kwargs)

    def request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> requests.Response:
        """Send a request to url.

        :attr:`.user_agent` is set as part of the headers for the request.
        All requests are logged through a debug logs, headers matching
        Authorization and Macaroons have their value replaced.

        :param method: HTTP method used for the request.
        :param url: URL to request with method.
        :param params: Query parameters to be sent along with the request.
        :param headers: Headers to be sent along with the request.

        :raises errors.StoreServerError: for error responses.
        :raises errors.NetworkError: for lower level network issues.

        :return: Response from the request.
        """
        if headers:
            headers["User-Agent"] = self.user_agent
        else:
            headers = {"User-Agent": self.user_agent}

        debug_headers = headers.copy()
        if debug_headers.get("Authorization"):
            debug_headers["Authorization"] = "<macaroon>"
        if debug_headers.get("Macaroons"):
            debug_headers["Macaroons"] = "<macaroon>"
        logger.debug(
            "HTTP %r for %r with params %r and headers %r",
            method,
            url,
            params,
            debug_headers,
        )
        try:
            response = self._session.request(
                method, url, headers=headers, params=params, **kwargs
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.RetryError,
        ) as error:
            raise errors.NetworkError(error) from error

        if not response.ok:
            raise errors.StoreServerError(response)

        return response



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Functions to serialize/deserialize credentials for Candid and Ubuntu One SSO."""

import json
from typing import Any, Dict, Literal

import pydantic
from pydantic import BaseModel, Field

from . import errors


class CandidModel(BaseModel):
    """Model for Candid credentials."""

    token_type: Literal["macaroon"] = Field("macaroon", alias="t")
    value: str = Field(..., alias="v")

    def marshal(self) -> Dict[str, Any]:
        """Create a dictionary containing the Candid credentials."""
        return self.dict(by_alias=True)

    @classmethod
    def unmarshal(cls, data: Dict[str, Any]) -> "CandidModel":
        """Create Candid model from dictionary data."""
        return cls(**data)


def marshal_candid_credentials(candid_creds: str) -> str:
    """Serialize Candid credentials for storage.

    This function creates a string that contains the desired `candid_creds` but also
    stores their "type", for unmarshalling later with `unmarshal_candid_credentials()`.

    :param candid_creds: The actual Candid credentials.
    :return: A payload string ready to be passed to Auth.set_credentials()
    """
    return json.dumps(CandidModel(v=candid_creds).marshal())  # pyright: ignore


def unmarshal_candid_credentials(marshalled_creds: str) -> str:
    """Deserialize previously stored Candid credentials.

    This function is meant to be called with the returned value from
    Auth.get_credentials(). As such, it supports parsing credentials from before we
    stored the `token_type`. Overall, this means supporting the scenarios where
    `marshalled_creds` is:

    (1) a regular JSON string generated by `marshal_candid_credentials()`;
    (2) a regular JSON string that was *not* generated by `marshal_candid_credentials()`
        but contains the Candid macaroon as a serialized dict;
    (3) a string that is not JSON but contains the serialized (non-dict) Candid macaroon.

    In addition, the function will raise a CredentialsNotParseable error when
    `marshalled_creds` is:

    (4) A regular JSON string containing a dict that *has* a type, but is *not* Candid.

    :param marshalled_creds: The credentials retrieved from auth storage.
    :return: The actual Candid credentials.
    """
    data: Dict[str, Any] = {}
    try:
        creds = json.loads(marshalled_creds)
    except json.JSONDecodeError:
        # Case (3): "opaque" serialized macaroon.
        data["v"] = marshalled_creds
    else:
        if "t" in creds:
            # Cases (1) and (4): dict containing the credentials type.
            data = creds
        else:
            # Case (2): dict without the type.
            data["v"] = marshalled_creds

    try:
        return CandidModel.unmarshal(data).value
    except pydantic.ValidationError as err:
        # Case (4): dict for some other credential type.
        raise errors.CredentialsNotParseable(
            "Expected valid Candid credentials"
        ) from err


class UbuntuOneMacaroons(BaseModel):
    """Model representation of the set of macaroons used in Ubuntu SSO."""

    root: str = Field(..., alias="r")
    discharge: str = Field(..., alias="d")

    def with_discharge(self, discharge: str) -> "UbuntuOneMacaroons":
        """Create a copy of this UbuntuOneMacaroons with a different discharge macaroon."""
        return self.copy(update={"d": discharge})


class UbuntuOneModel(BaseModel):
    """Model for Ubuntu One credentials."""

    token_type: Literal["u1-macaroon"] = Field("u1-macaroon", alias="t")
    value: UbuntuOneMacaroons = Field(..., alias="v")

    def marshal(self) -> Dict[str, Any]:
        """Create a dictionary containing the Ubuntu One credentials."""
        return self.dict(by_alias=True)

    @classmethod
    def unmarshal(cls, data: Dict[str, Any]) -> "UbuntuOneModel":
        """Create Candid model from dictionary data."""
        return cls(**data)


def marshal_u1_credentials(u1_creds: UbuntuOneMacaroons) -> str:
    """Serialize Ubuntu One credentials for storage.

    This function creates a string that contains the desired `u1_creds` but also
    stores their "type", for unmarshalling later with `unmarshal_u1_credentials()`.

    :param u1_creds: The actual Ubuntu One macaroons credentials.
    :return: A payload string ready to be passed to Auth.set_credentials()
    """
    return json.dumps(UbuntuOneModel(v=u1_creds).marshal())  # pyright: ignore


def unmarshal_u1_credentials(marshalled_creds: str) -> UbuntuOneMacaroons:
    """Deserialize previously stored Ubuntu One credentials.

    This function is meant to be called with the returned value from
    Auth.get_credentials(). As such, it supports parsing credentials from before we
    stored the `token_type`. Overall, this means supporting the scenarios where
    `marshalled_creds` is:

    (1) a regular JSON string generated by `marshal_u1_credentials()`;
    (2) a regular JSON string that was *not* generated by `marshal_u1_credentials()`
        but contains the `UbuntuOneMacaroons` values as a serialized dict;

    In addition, the function will raise a CredentialsNotParseable error when
    `marshalled_creds` is:

    (3) a string that is not JSON.
    (4) A regular JSON string containing a dict that *has* a type, but is *not* Ubuntu One.

    :param marshalled_creds: The credentials retrieved from auth storage.
    :return: The actual Ubuntu One credentials.
    """
    data: Dict[str, Any] = {}

    try:
        creds = json.loads(marshalled_creds)
    except json.JSONDecodeError as err:
        # Case (3).
        raise errors.CredentialsNotParseable(
            "Expected valid Ubuntu One credentials"
        ) from err
    else:
        if "t" in creds:
            # Cases (1) and (4).
            data = creds
        else:
            # Case (2).
            data["v"] = creds

    try:
        return UbuntuOneModel.unmarshal(data).value
    except pydantic.ValidationError as err:
        # Case (4): dict for some other credential type.
        raise errors.CredentialsNotParseable(
            "Expected valid Ubuntu One credentials"
        ) from err



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store BaseClient."""

import logging
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Sequence, cast
from urllib.parse import urlparse

import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from . import endpoints, errors, models
from .auth import Auth
from .http_client import HTTPClient

logger = logging.getLogger(__name__)


class BaseClient(metaclass=ABCMeta):
    """Encapsulates API calls for the Snap Store or Charmhub.

    :param base_url: the base url of the API endpoint.
    :param storage_base_url: the base url for storage.
    :param endpoints: :data:`.endpoints.CHARMHUB` or :data:`.endpoints.SNAP_STORE`.
    :param application_name: the name application using this class, used for the keyring.
    :param user_agent: User-Agent header to use for HTTP(s) requests.
    :param environment_auth: environment variable to use for credentials.
    :param ephemeral: keep everything in memory.

    :raises errors.NoKeyringError: if there is no usable keyring.
    """

    def __init__(
        self,
        *,
        base_url: str,
        storage_base_url: str,
        endpoints: endpoints.Endpoints,  # pylint: disable=W0621
        application_name: str,
        user_agent: str,
        environment_auth: Optional[str] = None,
        ephemeral: bool = False,
    ) -> None:
        """Initialize the Store Client."""
        self.http_client = HTTPClient(user_agent=user_agent)

        self._base_url = base_url
        self._storage_base_url = storage_base_url
        self._endpoints = endpoints

        self._auth = Auth(
            application_name,
            urlparse(base_url).netloc,
            environment_auth=environment_auth,
            ephemeral=ephemeral,
        )

    @abstractmethod
    def _get_discharged_macaroon(self, root_macaroon: str, **kwargs) -> str:
        """Return a discharged macaroon ready to use in an Authorization header."""

    @abstractmethod
    def _get_authorization_header(self) -> str:
        """Return the authorization header content to use."""

    def _get_macaroon(self, token_request: Dict[str, Any]) -> str:
        token_response = self.http_client.request(
            "POST",
            self._base_url + self._endpoints.tokens,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            json=token_request,
        )

        return token_response.json()["macaroon"]

    def login(
        self,
        *,
        permissions: Sequence[str],
        description: str,
        ttl: int,
        packages: Optional[Sequence[endpoints.Package]] = None,
        channels: Optional[Sequence[str]] = None,
        **kwargs,
    ) -> str:
        """Obtain credentials to perform authenticated requests.

        Credentials are stored on the system's keyring, handled by
        :data:`craft_store.auth.Auth`.

        The list of permissions to select from can be referred to on
        :data:`craft_store.attenuations`.

        The login process requires 3 steps:

        - request an initial macaroon on :attr:`.endpoints.Endpoints.tokens`.
        - discharge that macaroon using Candid
        - send the discharge macaroon to :attr:`.endpoints.Endpoints.tokens_exchange`
          to obtain final authorization of the macaroon

        This last macaroon is stored into the system's keyring to
        perform authenticated requests.

        :param permissions: Set of permissions to grant the login.
        :param description: Client description to refer to from the Store.
        :param ttl: time to live for the credential, in other words, how
                    long until it expires, expressed in seconds.
        :param packages: Sequence of packages to limit the credentials to.
        :param channels: Sequence of channel names to limit the credentials to.

        :raises errors.CredentialsAlreadyAvailable: if credentials already exist.
        """
        # Early check to ensure credentials do not already exist.
        self._auth.ensure_no_credentials()

        token_request = self._endpoints.get_token_request(
            permissions=permissions,
            description=description,
            ttl=ttl,
            packages=packages,
            channels=channels,
        )

        macaroon = self._get_macaroon(token_request)
        store_authorized_macaroon = self._get_discharged_macaroon(macaroon, **kwargs)

        # Save the authorization token.
        self._auth.set_credentials(store_authorized_macaroon)

        return self._auth.encode_credentials(store_authorized_macaroon)

    def request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> requests.Response:
        """Perform an authenticated request if auth_headers are True.

        :param method: HTTP method used for the request.
        :param url: URL to request with method.
        :param params: Query parameters to be sent along with the request.
        :param headers: Headers to be sent along with the request.

        :raises errors.StoreServerError: for error responses.
        :raises errors.NetworkError: for lower level network issues.
        :raises errors.CredentialsUnavailable: if credentials cannot be found.

        :return: Response from the request.
        """
        if headers is None:
            headers = {}

        headers["Authorization"] = self._get_authorization_header()

        return self.http_client.request(
            method,
            url,
            params=params,
            headers=headers,
            **kwargs,
        )

    def whoami(self) -> Dict[str, Any]:
        """Return whoami json data queyring :attr:`.endpoints.Endpoints.whoami`."""
        return self.request("GET", self._base_url + self._endpoints.whoami).json()

    def logout(self) -> None:
        """Clear credentials.

        :raises errors.CredentialsUnavailable: if credentials cannot be found.
        """
        self._auth.del_credentials()

    def upload_file(
        self,
        *,
        filepath: Path,
        monitor_callback: Optional[Callable] = None,
    ) -> str:
        """Upload filepath to storage.

        The monitor_callback is a method receiving one argument of type
        ``MultipartEncoder``, the total length of the upload can be accessed
        from this encoder from the ``len`` attribute to setup a progress bar
        instance.

        The callback is to return a function that receives a
        ``MultipartEncoderMonitor`` from which the ``.bytes_read`` attribute
        can be read to update progress.

        The simplest implementation can look like:

        .. code-block:: python

          def monitor_callback(encoder: requests_toolbelt.MultipartEncoder):

              # instantiate progress class with total bytes encoder.len

              def progress_printer(monitor: requests_toolbelt.MultipartEncoderMonitor):
                 # Print progress using monitor.bytes_read

              return progress_printer

        :param monitor_callback: a callback to monitor progress.

        """
        with filepath.open("rb") as upload_file:
            encoder = MultipartEncoder(
                fields={
                    "binary": (filepath.name, upload_file, "application/octet-stream")
                }
            )

            # create a monitor (so that progress can be displayed) as call the real pusher
            if monitor_callback is not None:
                monitor = MultipartEncoderMonitor(encoder, monitor_callback(encoder))
            else:
                monitor = MultipartEncoderMonitor(encoder)

            response = self.http_client.request(
                "POST",
                self._storage_base_url + self._endpoints.upload,
                headers={
                    "Content-Type": monitor.content_type,
                    "Accept": "application/json",
                },
                data=monitor,
            )

        result = response.json()
        if not result["successful"]:
            raise errors.CraftStoreError(f"Server error while pushing file: {result}")

        upload_id = self._endpoints.get_upload_id(result)
        logger.debug("Uploading bytes for %r ended, id %r", str(filepath), upload_id)

        return upload_id

    def notify_revision(
        self,
        *,
        name: str,
        revision_request: models.revisions_model.RevisionsRequestModel,
    ) -> models.revisions_model.RevisionsResponseModel:
        """Post to the revisions endpoint to notify the store about an upload.

        This request usually takes place after a successful :attr:`.upload`.
        """
        endpoint = f"/v1/{self._endpoints.namespace}/{name}/revisions"
        response = self.request(
            "POST", self._base_url + endpoint, json=revision_request.marshal()
        ).json()

        return cast(
            models.revisions_model.RevisionsResponseModel,
            models.revisions_model.RevisionsResponseModel.unmarshal(response),
        )

    def get_list_releases(self, *, name: str) -> models.MarshableModel:
        """Query the list_releases endpoint and return the result."""
        endpoint = f"/v1/{self._endpoints.namespace}/{name}/releases"
        response = self.request("GET", self._base_url + endpoint).json()

        return self._endpoints.list_releases_model.unmarshal(response)

    def release(
        self,
        *,
        name: str,
        release_request: Sequence[models.release_request_model.ReleaseRequestModel],
    ) -> None:
        """Request a release of name.

        :param name: name to release.
        :param release_request: sequence of items to release.
        """
        endpoint = f"/v1/{self._endpoints.namespace}/{name}/releases"

        self.request(
            "POST",
            self._base_url + endpoint,
            json=[r.marshal() for r in release_request],
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store StoreClient."""

import base64
import json
from typing import Optional

from macaroonbakery import bakery, httpbakery
from overrides import overrides
from pymacaroons.serializers import json_serializer

from . import creds, endpoints, errors
from .base_client import BaseClient
from .http_client import HTTPClient


def _macaroon_to_json_string(macaroon) -> str:
    return macaroon.serialize(json_serializer.JsonSerializer())


class WebBrowserWaitingInteractor(httpbakery.WebBrowserInteractor):
    """WebBrowserInteractor implementation using HTTPClient.

    Waiting for a token is implemented using HTTPClient which mounts
    a session with backoff retries.

    Better exception classes and messages are  provided to handle errors.
    """

    def __init__(self, user_agent: str) -> None:
        super().__init__()
        self.user_agent = user_agent

    # TODO: transfer implementation to macaroonbakery.
    def _wait_for_token(self, ctx, wait_token_url):
        request_client = HTTPClient(user_agent=self.user_agent)
        resp = request_client.request("GET", wait_token_url)
        if resp.status_code != 200:
            raise errors.CandidTokenTimeoutError(url=wait_token_url)
        json_resp = resp.json()
        kind = json_resp.get("kind")
        if kind is None:
            raise errors.CandidTokenKindError(url=wait_token_url)
        token_val = json_resp.get("token")
        if token_val is None:
            token_val = json_resp.get("token64")
            if token_val is None:
                raise errors.CandidTokenValueError(url=wait_token_url)
            token_val = base64.b64decode(token_val)
        return httpbakery._interactor.DischargeToken(  # pylint: disable=W0212
            kind=kind, value=token_val
        )


class StoreClient(BaseClient):
    """Encapsulates API calls for the Snap Store or Charmhub."""

    TOKEN_TYPE: str = "macaroon"

    @overrides
    def __init__(
        self,
        *,
        base_url: str,
        storage_base_url: str,
        endpoints: endpoints.Endpoints,  # pylint: disable=W0621
        application_name: str,
        user_agent: str,
        environment_auth: Optional[str] = None,
        ephemeral: bool = False,
    ) -> None:
        super().__init__(
            base_url=base_url,
            storage_base_url=storage_base_url,
            endpoints=endpoints,
            application_name=application_name,
            user_agent=user_agent,
            environment_auth=environment_auth,
            ephemeral=ephemeral,
        )

        self._bakery_client = httpbakery.Client(
            interaction_methods=[WebBrowserWaitingInteractor(user_agent=user_agent)]
        )

    def _get_authorization_header(self) -> str:
        auth = creds.unmarshal_candid_credentials(self._auth.get_credentials())

        return f"Macaroon {auth}"

    def _candid_discharge(self, macaroon: str) -> str:
        bakery_macaroon = bakery.Macaroon.from_dict(json.loads(macaroon))
        discharges = bakery.discharge_all(
            bakery_macaroon, self._bakery_client.acquire_discharge
        )

        # serialize macaroons the bakery-way
        discharged_macaroons = (
            "[" + ",".join(map(_macaroon_to_json_string, discharges)) + "]"
        )

        return base64.urlsafe_b64encode(discharged_macaroons.encode()).decode("ascii")

    def _authorize_token(self, candid_discharged_macaroon: str) -> str:
        token_exchange_response = self.http_client.request(
            "POST",
            self._base_url + self._endpoints.tokens_exchange,
            headers={"Macaroons": candid_discharged_macaroon},
            json={},
        )

        return token_exchange_response.json()["macaroon"]

    def _get_discharged_macaroon(self, root_macaroon: str, **kwargs) -> str:
        candid_discharged_macaroon = self._candid_discharge(root_macaroon)
        credentials = self._authorize_token(candid_discharged_macaroon)

        return creds.marshal_candid_credentials(credentials)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store StoreClient."""

from typing import Dict, Optional
from urllib.parse import urlparse

import requests
from overrides import overrides
from pymacaroons import Macaroon

from . import creds, endpoints, errors
from .base_client import BaseClient


class UbuntuOneStoreClient(BaseClient):
    """Encapsulates API calls for the Snap Store or Charmhub with Ubuntu One."""

    TOKEN_TYPE: str = "u1-macaroon"

    def __init__(
        self,
        *,
        base_url: str,
        storage_base_url: str,
        auth_url: str,
        endpoints: endpoints.Endpoints,  # pylint: disable=W0621
        application_name: str,
        user_agent: str,
        environment_auth: Optional[str] = None,
        ephemeral: bool = False,
    ) -> None:
        super().__init__(
            base_url=base_url,
            storage_base_url=storage_base_url,
            endpoints=endpoints,
            application_name=application_name,
            user_agent=user_agent,
            environment_auth=environment_auth,
            ephemeral=ephemeral,
        )
        self._auth_url = auth_url

    def _get_authorization_header(self) -> str:
        credentials = self._auth.get_credentials()
        macaroons = creds.unmarshal_u1_credentials(credentials)

        root_macaroon = Macaroon.deserialize(macaroons.root)
        discharged_macaroon = Macaroon.deserialize(macaroons.discharge)
        bound_macaroon = root_macaroon.prepare_for_request(
            discharged_macaroon
        ).serialize()
        return f"Macaroon root={macaroons.root}, discharge={bound_macaroon}"

    def _refresh_token(self) -> None:
        if self._endpoints.tokens_refresh is None:
            raise ValueError("tokens_refresh cannot be None")

        credentials = self._auth.get_credentials()
        macaroons = creds.unmarshal_u1_credentials(credentials)

        response = self.http_client.request(
            "POST",
            self._auth_url + self._endpoints.tokens_refresh,
            json={"discharge_macaroon": macaroons.discharge},
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if not response.ok:
            raise errors.StoreServerError(response)

        macaroons = macaroons.with_discharge(response.json()["discharge_macaroon"])

        new_credentials = creds.marshal_u1_credentials(macaroons)
        self._auth.set_credentials(new_credentials, force=True)

    def _extract_caveat_id(self, root_macaroon):
        macaroon = Macaroon.deserialize(root_macaroon)
        # macaroons are all bytes, never strings
        sso_host = urlparse(self._auth_url).netloc

        for caveat in macaroon.caveats:
            if caveat.location == sso_host:
                return caveat.caveat_id
        raise errors.CraftStoreError("Invalid root macaroon")

    def _discharge(
        self, email: str, password: str, otp: Optional[str], caveat_id
    ) -> str:
        data = dict(email=email, password=password, caveat_id=caveat_id)
        if otp:
            data["otp"] = otp

        response = self.http_client.request(
            "POST",
            self._auth_url + self._endpoints.tokens_exchange,
            json=data,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if not response.ok:
            raise errors.StoreServerError(response)

        return response.json()["discharge_macaroon"]

    def _get_discharged_macaroon(self, root_macaroon: str, **kwargs) -> str:
        email = kwargs["email"]
        password = kwargs["password"]
        otp = kwargs.get("otp")

        cavead_id = self._extract_caveat_id(root_macaroon)
        discharged_macaroon = self._discharge(
            email=email, password=password, otp=otp, caveat_id=cavead_id
        )

        u1_macaroon = creds.UbuntuOneMacaroons(r=root_macaroon, d=discharged_macaroon)
        return creds.marshal_u1_credentials(u1_macaroon)

    @overrides
    def request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> requests.Response:
        try:
            response = super().request(method, url, params, headers, **kwargs)
        except errors.StoreServerError as store_error:
            if "macaroon-needs-refresh" in store_error.error_list:
                self._refresh_token()
                response = super().request(method, url, params, headers, **kwargs)
            else:
                raise

        return response



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""List Releases Models for Charmhub responses."""

from datetime import datetime
from typing import Any, List, Optional

from ._base_model import MarshableModel
from ._common_list_releases_model import PackageModel, ProgressiveModel


class CharmBaseModel(MarshableModel):
    """Base entries for the channel-map entry from the list_releases endpoint."""

    architecture: str
    channel: str
    name: str


class ResourceModel(MarshableModel):
    """Resource entries for the channel-map entry from the list_releases endpoint."""

    name: str
    revision: Optional[int]
    type: str


class ChannelMapModel(MarshableModel):
    """Model for the channel-map results from the list_releases endpoint."""

    base: CharmBaseModel
    channel: str
    expiration_date: Optional[datetime]
    progressive: ProgressiveModel
    resources: List[ResourceModel]
    revision: int
    when: datetime


class RevisionModel(MarshableModel):
    """Model for a revision entry from list_releases."""

    bases: List[CharmBaseModel]
    created_at: datetime
    errors: Any
    revision: int
    sha3_384: str
    size: int
    status: str
    version: str


class ListReleasesModel(MarshableModel):
    """Model for the list_releases endpoint."""

    channel_map: List[ChannelMapModel]
    package: PackageModel
    revisions: List[RevisionModel]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store Authentication Store."""

import base64
import binascii
import logging
import os
from typing import Dict, Optional, Tuple

import keyring
import keyring.backend
import keyring.backends.fail
import keyring.errors

from . import errors

logger = logging.getLogger(__name__)


class MemoryKeyring(keyring.backend.KeyringBackend):
    """A keyring that stores credentials in a dictionary."""

    # Only > 0 make it to the chainer.
    priority = -1  # type: ignore

    def __init__(self) -> None:
        super().__init__()

        self._credentials: Dict[Tuple[str, str], str] = {}

    def set_password(self, service: str, username: str, password: str) -> None:
        """Set the service password for username in memory."""
        self._credentials[service, username] = password

    def get_password(self, service: str, username: str) -> Optional[str]:
        """Get the service password for username from memory."""
        return self._credentials.get((service, username))

    def delete_password(self, service: str, username: str) -> None:
        """Delete the service password for username from memory."""
        try:
            del self._credentials[service, username]
        except KeyError as key_error:
            raise keyring.errors.PasswordDeleteError() from key_error


class Auth:
    """Auth wraps around the keyring to store credentials.

    The application_name and host are used as key/values in the keyring to set,
    get and delete credentials.

    If environment_auth is set on initialization of this class, then a
    :attr:`MemoryKeyring` is setup in lieu of the system one.

    Credentials are base64 encoded into the keyring and decoded on
    retrieval.

    :ivar application_name: name of the application using this library.
    :ivar host: specific host for the store used.
    """

    def __init__(
        self,
        application_name: str,
        host: str,
        environment_auth: Optional[str] = None,
        ephemeral: bool = False,
    ) -> None:
        """Initialize Auth.

        :param application_name: name of the application using this library.
        :param host: specific host for the store used.
        :param environment_auth: environment variable used for authentication.
        :param ephemeral: keep everything in memory.
        """
        self.application_name = application_name
        self.host = host

        environment_auth_value = None
        if environment_auth:
            environment_auth_value = os.getenv(environment_auth)

        if environment_auth_value or ephemeral:
            keyring.set_keyring(MemoryKeyring())

        self._keyring = keyring.get_keyring()
        # This keyring would fail on first use, fail early instead.
        if isinstance(self._keyring, keyring.backends.fail.Keyring):
            raise errors.NoKeyringError()

        if environment_auth_value:
            self.set_credentials(self.decode_credentials(environment_auth_value))

    @staticmethod
    def decode_credentials(encoded_credentials: str) -> str:
        """Decode base64 encoded credentials.

        :raises errors.CredentialsNotParseable: when the credentials are incorrectly encoded.
        """
        try:
            return base64.b64decode(encoded_credentials).decode()
        except binascii.Error as err:
            raise errors.CredentialsNotParseable from err

    @staticmethod
    def encode_credentials(credentials: str) -> str:
        """Encode credentials to base64."""
        return base64.b64encode(credentials.encode()).decode()

    def ensure_no_credentials(self) -> None:
        """Check that no credentials exist.

        :raises errors.CredentialsAvailable: if credentials have already been set.
        """
        if self._keyring.get_password(self.application_name, self.host) is not None:
            raise errors.CredentialsAlreadyAvailable(self.application_name, self.host)

    def set_credentials(self, credentials: str, force: bool = False) -> None:
        """Store credentials in the keyring.

        :param credentials: token to store.
        :param force: overwrite existing credentials.
        """
        if not force:
            self.ensure_no_credentials()

        logger.debug(
            "Storing credentials for %r on %r in keyring %r.",
            self.application_name,
            self.host,
            self._keyring.name,
        )
        encoded_credentials = self.encode_credentials(credentials)
        self._keyring.set_password(
            self.application_name, self.host, encoded_credentials
        )

    def get_credentials(self) -> str:
        """Retrieve credentials from the keyring."""
        logger.debug(
            "Retrieving credentials for %r on %r from keyring %r.",
            self.application_name,
            self.host,
            self._keyring.name,
        )

        try:
            encoded_credentials_string = self._keyring.get_password(
                self.application_name, self.host
            )
        except Exception as unknown_error:
            logger.debug(
                "Unhandled exception raised when retrieving credentials: %r",
                unknown_error,
            )
            raise errors.CredentialsUnavailable(
                self.application_name, self.host
            ) from unknown_error

        if encoded_credentials_string is None:
            logger.debug("Credentials not found in the keyring %r", self._keyring.name)
            raise errors.CredentialsUnavailable(self.application_name, self.host)
        credentials = self.decode_credentials(encoded_credentials_string)
        return credentials

    def del_credentials(self) -> None:
        """Delete credentials from the keyring."""
        # Try to get the credentials first to see if there are any,
        # this is to provide an easier troubleshooting experience.
        self.get_credentials()

        logger.debug(
            "Deleting credentials for %r on %r from keyring %r.",
            self.application_name,
            self.host,
            self._keyring.name,
        )
        self._keyring.delete_password(self.application_name, self.host)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Common Models between namespaces for List Releases responses."""


from typing import List, Optional

from ._base_model import MarshableModel


class ProgressiveModel(MarshableModel):
    """Model for the progressive information from the channel-map model.

    :param paused: signals if the progressive release is paused on a channel.
    :param percentage: the progress of a progressive release on a channel.
    """

    paused: Optional[bool]
    percentage: Optional[float]


class ChannelsModel(MarshableModel):
    """Model for the channels results from the list_releases endpoint.

    :param branch: the channel branch.
    :param fallback: the channel to fallback to if this one is closed.
    :param name: the full name of the channel (<track>/<risk>[/<branch>]).
    :param risk: the channel risk (one of stable, candidate, beta or edge).
    :param track: the channel track.
    """

    branch: Optional[str]
    fallback: Optional[str]
    name: str
    risk: str
    track: str


class PackageModel(MarshableModel):
    """Model for the package results from the list_releases endpoint.

    :param channels: list of :attr:`ChannelsModel`.
    """

    channels: List[ChannelsModel]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Release request models for the Store."""

from typing import List, Optional

from pydantic import Field

from ._base_model import MarshableModel


class ResourceModel(MarshableModel):
    """Resource model for a ReleaseRequestModel.

    :param name: name of the resource to release.
    :param revision: revision of the resource to release.
    """

    name: str
    revision: Optional[int]


class ReleaseRequestModel(MarshableModel):
    """Model to request a release to the store.

    :param channel: name of the channel to release to.
    :param resources: resources to release with this revision.
    :param revision: revision to release.
    """

    channel: str
    resources: Optional[List[ResourceModel]] = Field(default_factory=list)
    revision: Optional[int] = Field(...)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Craft Store errors."""

import contextlib
import logging
from typing import Dict, List, Optional

import requests
import urllib3  # type: ignore
from requests.exceptions import JSONDecodeError

logger = logging.getLogger(__name__)


class CraftStoreError(Exception):
    """Base class error for craft-store."""

    def __init__(self, message: str, resolution: Optional[str] = None) -> None:
        super().__init__(message)
        self.resolution = resolution


class NetworkError(CraftStoreError):
    """Error to raise on network or infrastructure issues.

    The original exception is used to potentially craft a user friendly
    error message to be used for :attr:`.brief`.

    :param exception: original exception raised.

    :ivar exception: original exception raised.
    """

    def __init__(self, exception: Exception) -> None:
        message = str(exception)
        with contextlib.suppress(IndexError):
            if isinstance(exception.args[0], urllib3.exceptions.MaxRetryError):  # type: ignore
                message = "Maximum retries exceeded trying to reach the store."

        super().__init__(message)


class StoreErrorList:
    """Error List returned from the Store."""

    def __len__(self) -> int:
        return len(self._error_list)

    def __str__(self) -> str:
        error_list: List[str] = []
        for error in self._error_list:
            error_list.append(f"- {error['code']}: {error['message']}")
        return "\n".join(error_list).strip()

    def __repr__(self) -> str:
        code_list = []
        for error in self._error_list:
            code = error.get("code")
            if code:
                code_list.append(code)

        return "<StoreErrorList: {' '.join(code_list)}>"

    def __contains__(self, error_code: str) -> bool:
        return any((error.get("code") == error_code for error in self._error_list))

    def __getitem__(self, error_code: str) -> Dict[str, str]:
        for error in self._error_list:
            if error.get("code") == error_code:
                return error

        raise KeyError(error_code)

    def __init__(self, error_list: List[Dict[str, str]]) -> None:
        self._error_list = error_list


class StoreServerError(CraftStoreError):
    """Error to raise on infrastructure issues from error codes above ``500``.

    :param response: the response from a :class:`requests.Request`.

    :ivar response: the response from a :class:`requests.Request`.
    :ivar error_list: list of errors returned by the Store :class:`StoreErrorList`.
    """

    def _get_raw_error_list(self) -> List[Dict[str, str]]:
        response_json = self.response.json()
        try:
            # Charmhub uses error-list.
            error_list = response_json["error-list"]
        except KeyError:
            # Snap Store uses error_list.
            error_list = response_json["error_list"]

        return error_list

    def __init__(self, response: requests.Response) -> None:
        self.response = response

        try:
            raw_error_list: List[Dict[str, str]] = self._get_raw_error_list()
        except (KeyError, JSONDecodeError):
            raw_error_list = []

        self.error_list = StoreErrorList(raw_error_list)

        message: Optional[str] = None
        if self.error_list:
            with contextlib.suppress(KeyError):
                message = "Store operation failed:\n" + str(self.error_list)
        if message is None:
            message = (
                "Issue encountered while processing your request: "
                f"[{response.status_code}] {response.reason}."
            )

        super().__init__(message)


class CredentialsAlreadyAvailable(CraftStoreError):
    """Error raised when credentials are already found in the keyring."""

    def __init__(self, application: str, host: str) -> None:
        super().__init__(f"Credentials found for {application!r} on {host!r}.")


class CredentialsUnavailable(CraftStoreError):
    """Error raised when credentials are not found in the keyring."""

    def __init__(self, application: str, host: str) -> None:
        super().__init__(f"No credentials found for {application!r} on {host!r}.")


class CredentialsNotParseable(CraftStoreError):
    """Error raised when credentials are not parseable."""

    def __init__(self, msg: str = "Expected base64 encoded credentials") -> None:
        super().__init__(f"Credentials could not be parsed. {msg}.")


class NoKeyringError(CraftStoreError):
    """Error raised when no keyring can be used."""

    def __init__(self) -> None:
        super().__init__("No keyring found to store or retrieve credentials from.")


class CandidTokenTimeoutError(CraftStoreError):
    """Error raised when timeout is reached trying to discharge a macaroon."""

    def __init__(self, url: str) -> None:
        super().__init__(f"Timed out waiting for token response from {url!r}.")


class CandidTokenKindError(CraftStoreError):
    """Error raised when the token kind is missing from the discharged macaroon."""

    def __init__(self, url: str) -> None:
        super().__init__(f"Empty token kind returned from {url!r}.")


class CandidTokenValueError(CraftStoreError):
    """Error raised when the token value is missing from the discharged macaroon."""

    def __init__(self, url: str) -> None:
        super().__init__(f"Empty token value returned from {url!r}.")
